/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: evaluate.c,v 1.2 1999/11/06 01:07:49 shinmaru Exp $
 */

#include "mahjong.h"
#include "hand.h"
#include "option.h"

static int __pid;
static int __last_tile;
static int __min_going;
static int __revealed;
static int __one_suit;
static int __forbids;
static int __self_drawn;

going_table_t going_table[MAX_SORT];

static int record_singleton(int n, evaluated_hand_t *e, hook_table_t h);
static int record_going_out(evaluated_hand_t *e);
static int record_going_out_7_pairs(evaluated_hand_t *e);
static int record_going_out_13_orphans(evaluated_hand_t *e);
static void compose(set_t *sets, int count, int going);
static int correct_going_table(int *hand);
static void evaluate_7_pairs(int *hand);
static void evaluate_13_orphans(int *hand);

typedef int (*estimator_t)(evaluated_hand_t *e);
static estimator_t estimators[MAX_EVAL] = {
  estimate_no_terminals,
  estimate_terminals,
  estimate_sequences,
  estimate_sisters,
  estimate_straight,
  estimate_3_suit_equals,
  estimate_1_suit,
  estimate_triplets,
  estimate_7_pairs,
  estimate_13_orphans
};

static int update_scoresheet(int pid, scoresheet_t *sheet)
{
  static scoresheet_t __scoresheet[MAX_PLAYER];
  player_t *p = &players[pid];
  if (sheet != NULL) {
    if (p->scoresheet == NULL ||
        p->scoresheet->total < sheet->total ||
        p->scoresheet->bonus < sheet->bonus) {
      __scoresheet[pid] = *sheet; /* Τ˥ԡ */
      p->scoresheet = &__scoresheet[pid];
      return 1;
    }
  }
  return p->scoresheet != NULL;
}

static void receiver(set_t *sets, int count)
{
  int i;
  int going = calculate_going(sets, count);
  if (going > __min_going) return;
  if (going < __min_going)
    __min_going = going;

  for (i = 0; i < players[__pid].revealed_count; i++) {
    sets[count++] = players[__pid].revealed[i];
  }
#ifdef __DEBUG__  
  if (count >= MAX_HANDED_SET) {
    fprintf(stderr, "evaluate#receiver(): Buffer overflowed\n");
    exit(1); /*  */
  }
  if (!option_silence) {
    int i;
    for (i = 0; i < count; i++) {
      print_set(sets[i]);
    }
    printf(": %d\n", going);
  }
#endif
  compose(sets, count, going);
}

/* ǽפꤷʤ硢nȤϤ
 * ξλϼʤ */
int evaluate(int pid, int n)
{
  player_t *p = &players[pid];  
  int i, count;

  for (i = 0, count = 0; i < MAX_SORT; i++) {
    count += p->hand[i];
  }
  if (count > 14) {
    fprintf(stderr, "evaluate(): Too many hand.\n");
    exit(1); /* ¿ */
  }

  memset(going_table, 0, sizeof(going_table));
  p->scoresheet = NULL;
  
  __pid = pid;
  __self_drawn = pid == current_discarder();
  __last_tile = n;
  __min_going = INT_MAX;
  __forbids = bit(EVAL_7_PAIRS) | bit(EVAL_13_ORPHANS);
  __revealed = bittest(p->condition, COND_REVEALED);

  if (__revealed) {
    /* ϪˤäΩʤɾػߤ */
    __forbids |= bit(EVAL_SEQUENCES);
    __forbids |= bit(EVAL_SISTERS);
  }

  /* Ϫ̻ҤοĴ٤ */
  __one_suit = most_suit(pid);
  if (__one_suit < 0) {
    __forbids |= bit(EVAL_1_SUIT);
  }

  generate_combinations(p->hand, receiver); /* ̷ɾ */
  if (p->revealed_count == 0) {
    evaluate_7_pairs(p->hand); /* лҤɾ */
    evaluate_13_orphans(p->hand); /* ̵Фɾ */
  }
  
  return correct_going_table(p->hand);
}

int most_suit(int pid)
{
  int characters = 0, balls = 0, bamboos = 0;
  int total;
  int i;
  int revealed_suits = 0;
  player_t *p = &players[pid];

  for (i = 0; i < p->revealed_count; i++) {
    revealed_suits |= bit(suit_of(tile_of(p->revealed[i])));
  }
  revealed_suits &= ~bit(SUIT_HONOR);
  switch (revealed_suits) {
  case bit(SUIT_CHARACTER): return SUIT_CHARACTER;
  case bit(SUIT_BALL):      return SUIT_BALL;
  case bit(SUIT_BAMBOO):    return SUIT_BAMBOO;
  case 0:                   break;
  default:		    return -1;
  }

  i = 0;
  for (; i <  9; i++) characters += p->hand[i];
  for (; i < 18; i++) balls += p->hand[i];
  for (; i < 27; i++) bamboos += p->hand[i];

  total = (characters + balls + bamboos) / 2;

  if (characters > total) return SUIT_CHARACTER;
  if (balls > total) return SUIT_BALL;
  if (bamboos > total) return SUIT_BAMBOO;
  return -1;
}

int correct_going_table(int *hand)
{
  int max_gone;
  int n;
  for (n = 0; n < MAX_SORT; n++) {
    if (hand[n] == 4) {
      /* פǻȤڤäƤפǤԤ̵ˤ */      
      int i, j;
      for (i = 0; i < MAX_SORT; i++) {
        going_table_t *g = &going_table[i];
        int types;        
        if (g->hook[n].types != 0) {
          g->hook[n].types = 0;
          for (j = 0, types = 0; j < MAX_SORT; j++) {
            types |= g->hook[j].types;
          }
          if (types == 0)
            g->max_gone -= 1;
        }
      }
    }
  }
  max_gone = 0;
  for (n = 0; n < MAX_SORT; n++) {
    if (max_gone < going_table[n].max_gone)
      max_gone = going_table[n].max_gone;
  }
  if (max_gone == 0)
    /* λʤ
     * Ωʤλл˿ʤΤǼºݤˤͭʤ */
    return 0;
  else
    return 14 - max_gone;
}

static int subdivide_immature_set(evaluated_hand_t *e, int i)
{
  set_t set = e->sets[i];
  int n1 = tile_of(set);
  int n2;
  switch (type_of(set)) {
  case PAIR:
    n2 = n1;
    e->triplets[n1] -= 2;
    break;
  case CONJUNCT:
    n2 = n1 + 1;
          if (number_of(n1) == 7) e->sequences[n1 - 1] -= 2;
    else  if (number_of(n1) == 0) e->sequences[n1    ] -= 2;
    else { e->sequences[n1] -= 1; e->sequences[n1 - 1] -= 1; }
    break;
  case SPLITTER:
    n2 = n1 + 2;
    e->sequences[n1] -= 2;
    break;
  default:
    fprintf(stderr, "subdivide_immature_set(): can't divide set\n");
    exit(1);
  }
  e->sets[i         ] = make_set(n1, SINGLETON);
  e->sets[e->count++] = make_set(n2, SINGLETON);
  e->singletons[n1]++;
  e->singletons[n2]++;
  return n2;
}

static void repair(evaluated_hand_t *e, int i)
{
  int n1 = tile_of(e->sets[i]);
  int n2 = tile_of(e->sets[--e->count]);
  switch (n2 - n1) {
  case 0:
    e->sets[i] = make_set(n1, PAIR);
    e->triplets[n1] += 2;
    break;
  case 1:
    e->sets[i] = make_set(n1, CONJUNCT);
         if (number_of(n1) == 7)  e->sequences[n1 - 1] += 2;
    else if (number_of(n1) == 0)  e->sequences[n1    ] += 2;
    else { e->sequences[n1] += 1; e->sequences[n1 - 1] += 1; }
    break;
  case 2:
    e->sets[i] = make_set(n1, SPLITTER);
    e->sequences[n1] += 2;
    break;
  default:
    fprintf(stderr, "repair(): can't repair set\n");
    exit(1);
  }
  e->singletons[n1]--;
  e->singletons[n2]--;
}

void compose(set_t *sets, int count, int going)
{
  int number_of_singletons = 0, number_of_pairs = 0;
  int last_singleton;
  int i;
  hook_table_t h;
  evaluated_hand_t e;

  e.pid = __pid;
  e.sets = sets;
  e.count = count;
  e.going = going;
  e.revealed = __revealed;
  e.one_suit = __one_suit;
  e.forbids = __forbids;
  for (i = 0; i < MAX_SORT; i++)
    e.singletons[i] = e.sequences[i] = e.triplets[i] = 0;

  for (i = 0; i < count; i++) {
    set_t set = sets[i];
    int n = tile_of(set);
    switch (type_of(set)) {
    case SINGLETON:
      e.singletons[n] += 1;
      last_singleton = n;
      number_of_singletons++;
      break;
    case PAIR:
      e.triplets[n] += 2;
      number_of_pairs++;
      break;
    case CONJUNCT:
           if (number_of(n) == 7) e.sequences[n - 1] += 2;
      else if (number_of(n) == 0) e.sequences[n    ] += 2;
      else { e.sequences[n] += 1; e.sequences[n - 1] += 1; }
      break;
    case SPLITTER:
      e.sequences[n] += 2;
      break;
    case SEQUENCE:
      e.sequences[n] += 3;
      e.forbids |= bit(EVAL_TRIPLETS);
      break;
    case TRIPLET:
    case QUADRUPLET:
      e.triplets[n] += 3; /* ҤȤư */
      e.forbids |= bit(EVAL_SEQUENCES);
      e.forbids |= bit(is_terminal(set) ? EVAL_NO_TERMINALS : EVAL_TERMINALS);
      break;
    }
  }

  if (number_of_singletons > 0) {
    /* ƸΩפˤĤפɾ */
    memset(h, 0, sizeof(h));
    detect_hooks(sets, count, e.going, h);
    
    for (i = 0; i < count; i++) {
      int n = tile_of(sets[i]);
      if (type_of(sets[i]) == SINGLETON) {
        record_singleton(n, &e, h);
#ifdef __DEBUG__
        printf("Ω %s\n", tile_face[n]);
#endif
      }
    }
  }
  if (number_of_singletons == 1 && number_of_pairs == 1 && e.going == 1) {
    int n = last_singleton;
    if (is_numeric(n) && 2 <= number_of(n) && number_of(n) <= 6) {
      /* 뤬ʤСפĤθ */
      memset(h, 0, sizeof(h));
      for (i = 0; i < count; i++) {
        int n1 = tile_of(sets[i]);
        int n2;
        switch (type_of(sets[i])) {
        case CONJUNCT:
          /* ξڤʤ */
          if (number_of(n1) != 0 && number_of(n1) != 7)
            goto done;
          /* ȴ */
        case SPLITTER:
          /* ϸİ1䤹
           * ιԤθƽФǤϼºݤ˸İʤ */
#ifdef __DEBUG__
          printf(":");
          print_set(sets[i]);
          putchar('\n');
#endif
          n2 = subdivide_immature_set(&e, i);
          e.going += 1;
          detect_hooks(sets, e.count, e.going, h);
          record_singleton(n1, &e, h);
          record_singleton(n2, &e, h);
          e.going -= 1;
          repair(&e, i);
          goto done;
        }
      }
    done:
    }
  }
  if (number_of_singletons == 0) {
    int immaturities[MAX_SORT];
    int bits;
    if (e.going == 0) {
      if (record_going_out(&e)) return; /* λ */
      e.going++; /* λʤпƬȤˤ */
      number_of_pairs = 0;
    }
    /* ƬȸİΤͣлҤϲ¸ */
    bits = ~((number_of_pairs == 1 ? 0 : bit(PAIR)) | bit(CONJUNCT) | bit(SPLITTER));

    for (i = 0; i < MAX_SORT; i++) immaturities[i] = bits;
    for (i = 0; i < count; i++) {
      int type = type_of(sets[i]);
      int n1 = tile_of(sets[i]);
      int n2;
      if (!bittest(immaturities[n1], type)) {
#ifdef __DEBUG__
        printf(":");
        print_set(sets[i]);
        putchar('\n');
#endif
        immaturities[n1] |= bit(type);
        n2 = subdivide_immature_set(&e, i);
        memset(h, 0, sizeof(h));
        detect_hooks(sets, e.count, e.going, h);
        record_singleton(n1, &e, h);
        if (n1 != n2)
          record_singleton(n2, &e, h);
        repair(&e, i);
      }
    }
  }
}

void evaluate_7_pairs(int *hand)
{
  set_t sets[MAX_HANDED_SET], *s;
  hook_table_t h;
  evaluated_hand_t e;
  int triplets = 0, going;
  int i;

  memset(h, 0, sizeof(h));
  memset(&e, 0, sizeof(e));

  going = calculate_going_7_pairs(hand, h);
  if (going > __min_going) return;
  if (going < __min_going)
    __min_going = going;

  for (i = 0, s = sets; i < MAX_SORT; i++) {
    if (hand[i] >= 2) {
      int n = hand[i] - 2;
      e.triplets[i] += 2;
      e.singletons[i] += n;
      *s++ = make_set(i, PAIR);
      if (n > 0) triplets++;
      while (n-- > 0)
        *s++ = make_set(i, SINGLETON);
    } else if (hand[i] == 1) {
      e.singletons[i]++;
      *s++ = make_set(i, SINGLETON);
    }
  }

  e.pid = __pid;
  e.sets = sets;
  e.count = s - sets;
  e.going = going;
  e.one_suit = __one_suit;
  e.forbids = ~(bit(EVAL_NO_TERMINALS) | bit(EVAL_TERMINALS) | bit(EVAL_1_SUIT) |
                bit(EVAL_7_PAIRS));
  
  if (going == 0) {
    if (record_going_out_7_pairs(&e))
      return;
  }
  /* Ū˹ҤǤäƤڤäƤС¤Ȥξŷˤ硢
   * (̷)Сθ褦פǽǤ */
  if (going == 1 && triplets == 1) {
    /* [3/1/2/2/2/2/2]İפξ硢Ҥʤиİ1 */
    for (i = 0; i < MAX_SORT; i++) {
      if (hand[i] == 3) {
        record_singleton(i, &e, h);
      } else if (hand[i] == 1) {
        e.going += 1;
        record_singleton(i, &e, h);
        e.going -= 1;
      }
    }
  } else {
    for (i = 0; i < MAX_SORT; i++) {
      if (e.singletons[i])
        record_singleton(i, &e, h);
    }
  }
}

void evaluate_13_orphans(int *hand)
{
  set_t sets[MAX_HANDED_SET], *s;
  hook_table_t h;
  evaluated_hand_t e;
  int no_uses, going;
  int i;

  memset(h, 0, sizeof(h));
  memset(&e, 0, sizeof(e));

  going = calculate_going_13_orphans(hand, h);

  if (going > __min_going) return;
  if (going < __min_going)
    __min_going = going;

  no_uses = 0;
  for (i = 0, s = sets; i < MAX_SORT; i++) {
    if (hand[i]) {
      int n = hand[i];
      while (n-- > 0) {
        *s++ = make_set(i, SINGLETON);
      }
      e.singletons[i] = hand[i];
      if (!is_honor(i) && number_of(i) != 0 && number_of(i) != 8)
        no_uses++;
    }
  }

  e.pid = __pid;
  e.sets = sets;
  e.count = s - sets;
  e.going = going;
  e.one_suit = __one_suit;
  e.forbids = ~bit(EVAL_13_ORPHANS);
  
  if (going == 0) {
    if (record_going_out_13_orphans(&e))
      return;
  }
  for (i = 0; i < MAX_SORT; i++) {
    if (hand[i]) {
      if (is_honor(i) || number_of(i) == 0 || number_of(i) == 8) {
        /* ĥפڤꤪޤϷƬפ¸ */
        if (hand[i] == 1 || no_uses != 0)
          continue;
      }
      record_singleton(i, &e, h);
    }
  }
}

int record_singleton(int n, evaluated_hand_t *e, hook_table_t h)
{
  going_table_t *g = &going_table[n];
  int gone = 14 - e->going;
  int i;
  
  /* ΩפñԤʳǴԤƤʤСİ */
  if ((h[n].types & ~bit(HOOK_PAIR)) != 0)
    return 0;

  if (g->max_gone > gone) return 0;
  if (g->max_gone < gone) {
    /* ɽƤ쿷 */
    g->max_gone = gone;
    g->quadrupling = 0;
    if (e->one_suit >= 0) {
      g->one_suit = bit(e->one_suit);
    }
    memcpy(g->hook, h, sizeof(hook_table_t));
    for (i = 0; i < MAX_EVAL; i++) {
      g->values[i] = 0;
    }
  } else {
    /* Ωפ¸ԤפʤСİ */
    if (g->hook[n].types != 0)
      return 0;
    /* ¸ͭפʻ礹 */
    if (e->one_suit >= 0) {
      g->one_suit |= bit(e->one_suit);
    }
    for (i = 0; i < MAX_SORT; i++) {
      if (h[i].needs > 0) {
        g->hook[i].needs = max(h[i].needs, g->hook[i].needs);
        g->hook[i].types |= h[i].types;
      }
    }
  }
  g->hook[n].needs = 0;
  g->hook[n].types = 0;
  g->quadrupling |= e->triplets[n] >= 3;

  /* ɾ */
  e->singletons[n]--;
  for (i = 0; i < MAX_EVAL; i++) {
    if (!bittest(e->forbids, i)) {
      int value = (*estimators[i])(e);
      g->values[i] = max(g->values[i], value);
    }
  }
  e->singletons[n]++;
  return 1;
}

int record_going_out(evaluated_hand_t *e)
{
  int point = 20, honors = 0, prizes;
  int hands, completions;
  int hook_types, n;
  int map[MAX_SORT], values[MAX_EVAL];
  int i;
  set_t sets[MAX_HANDED_SET], *original_sets;
  scoresheet_t *sheet;

  if (__last_tile < 0)
    return 0; /* ǽפꤵƤʤ */

  hook_types = players[e->pid].winnings[__last_tile].types;
  if (hook_types == 0) {
#if 0
    fprintf(stderr, "record_going_out(): Tell me how to go out with %s.\n",
            tile_face[__last_tile]);
    return 0; /* Ԥηʬʤ */
#else
    hook_types = find_hook_types(e->sets, e->count, __last_tile);
#endif
  }
  
  /* ¤ξ硢ǽפΩҤϰŹǤϤʤ
   * ؤԤɬͥ褹 */
  if (!__self_drawn && (hook_types & ~bit(HOOK_TRIPLET)) != 0) {
    hook_types &= ~bit(HOOK_TRIPLET);
  }

  for (i = 0; i < MAX_SORT; i++) map[i] = 0;
  for (i = 0; i < e->count; i++) {
    set_t set = e->sets[i];
    int n = tile_of(set);
    switch (type_of(set)) {
    case SINGLETON: map[n] += 1; break;
    case PAIR: map[n] += 2; break;
    case TRIPLET:
      map[n] += 3;
      if (!__self_drawn && n == __last_tile && hook_types == bit(HOOK_TRIPLET)) {
        /* ¤ˤäƴҤ */
        set = toggle_revealed_bit(set);
      }
      break;
    case QUADRUPLET:
      map[n] += 4;
      if (mark_of(set) == 0) {
	/* ʤ */
	set = make_set(n, QUADRUPLET);
      }
      break;
    case SEQUENCE: map[n]++; map[n + 1]++; map[n + 2]++; break;
    default:
      fprintf(stderr, "record_going_out(): Illegal hand.\n");
      exit(1);
    }
    point += point_of(set, e->pid);
    sets[i] = set;
  }

  for (i = 27; i < 30; i++) {
    if (map[i] >= 3) honors++;
  }
  if (map[round_wind()] >= 3) honors++;
  if (map[seat_wind_of(e->pid)] >= 3) honors++;

  /* ȽǤ */
  original_sets = e->sets;
  e->sets = sets;
  
  for (i = 0; i < MAX_EVAL; i++) {
    values[i] = bittest(e->forbids, i) ? 0 : (*estimators[i])(e);
  }
  hands = judge(map, values, e);
  completions = tell_completion_bonus(e);

  e->sets = original_sets;

  /* ǽפԤꤹ */
  n = __last_tile;
  if (bittest(hands, ALL_SEQUENCES)) {
    if ((bittest(hook_types, HOOK_BELOW) && number_of(n) != 6) ||
        (bittest(hook_types, HOOK_ABOVE) && number_of(n) != 2)) {
      /* ξԤ˸¤ */
      hook_types = 0;
    } else {
      /* ξԤʳʿ¤Ωʤ */
      hands &= ~bit(ALL_SEQUENCES);
    }
  }

  if (hands == 0 && honors == 0 && completions == 0)
    return 0; /* ʤ */

  /* ξԤˤԤĤʤΤǥե饰򲼤
   * äƤɤ餫Υե饰ΩäƤʤĥԤǤ */
  switch (number_of(n)) {
  case 2: hook_types &= ~bit(HOOK_BELOW); break;
  case 6: hook_types &= ~bit(HOOK_ABOVE); break;
  default:
    hook_types &= ~(bit(HOOK_BELOW) | bit(HOOK_ABOVE));
  }
  if (hook_types != 0) {
    /* Ԥ */    
    point += 2;
  }
  if (__self_drawn) {
    /*  */
    point += bittest(hands, ALL_SEQUENCES) ? 0 : 2;
  }
  else if (!__revealed) {
    /*  */
    point += 10;
  }
  
  point = (point + 9) / 10 * 10;
  prizes = count_prizes(map, is_ready(e->pid)); /* ɥ */
  sheet = compute_score(e->pid, hands, completions, honors, prizes, point, __revealed);
  return update_scoresheet(e->pid, sheet);
}

int record_going_out_7_pairs(evaluated_hand_t *e)
{
  int point = 25, prizes;
  int hands, completions;
  int values[MAX_EVAL];
  int i;
  scoresheet_t *sheet;

  for (i = 0; i < MAX_EVAL; i++) {
    values[i] = bittest(e->forbids, i) ? 0 : (*estimators[i])(e);
  }
  hands = judge(e->triplets, values, e);
  completions = tell_completion_bonus(e);

  if (hands == 0 && completions == 0)
    return 0; /* ʤ */

  prizes = count_prizes(e->triplets, is_ready(e->pid)); /* ɥ */
  sheet = compute_score(e->pid, hands, completions, 0, prizes, point, 0);
  return update_scoresheet(e->pid, sheet);
}

int record_going_out_13_orphans(evaluated_hand_t *e)
{
  int hands, completions;
  scoresheet_t *sheet;
  
  hands = bit(THIRTEEN_ORPHANS);
  completions = tell_completion_bonus(e);

  if (hands == 0 && completions == 0)
    return 0; /* ʤ */

  sheet = compute_score(e->pid, hands, completions, 0, 0, 0, 0);
  return update_scoresheet(e->pid, sheet);
}
