/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: pool.c,v 1.2 1999/11/06 01:07:50 shinmaru Exp $
 */

#include "mahjong.h"
#include "debug.h"

#define MAX_POOL		136

static int *pool_head, *pool_tail, *pool_tomb;
static int pool[MAX_POOL];
static int * const prize_marking = &pool[MAX_POOL - 6];

bone_t *wall_bones;

int pool_finished(void)
{
  return pool_head >= pool_tomb;
}

int pool_remain(void)
{
  return max(0, (int)(pool_tomb - pool_head));
}

int pool_count_quadruplets(void)
{
  return &pool[MAX_POOL] - pool_tail;
}

/* פΤäǸƤӽФ */
static void open_prize_marker(void)
{
  /*   
   * ca86420 
   * db97531 
   *---------
   */
  int position;
  int i, n;
  bone_t *b;

  position = 4 + pool_count_quadruplets(); /* פŪʰ */

  for (b = wall_bones; b != NULL; b = b->next) {
    if (position-- == 0) {
      n = tile_of(b->tid);
      b->tid = make_tid(n, 0, SURFACE_FACE);
      break;
    }
  }

  for (i = 0; i < MAX_PLAYER; i++)
    players[i].dead_tiles[n]++;
}

void pool_shuffle(void)
{
  int temp[MAX_POOL];
  int i;

  for (i = 0; i < MAX_PLAYER; i++) {
    int n;
    for (n = 0; n < MAX_SORT; n++) {
      players[i].hand[n] = players[i].dead_tiles[n] = 0;
    }
  }

  for (i = 0; i < MAX_POOL; i++) {
    temp[i] = i / 4;
  } 

  for (i = 0; i < MAX_POOL; i++) {
    int p = random_integer(MAX_POOL - i);
    pool[i] = temp[p];
    temp[p] = temp[MAX_POOL - i - 1];
  }
  pool_head = pool;
  pool_tail = pool + MAX_POOL;
  pool_tomb = pool + MAX_POOL - 14;
}

void pool_deal(void)
{
  int i, n;
  for (n = 0; n < 3; n++) {
    for (i = 0; i < MAX_PLAYER; i++) {
      int p = (current_dealer() + i) % 4;
      pool_draw(p);
      pool_draw(p);
      pool_draw(p);
      pool_draw(p);
    }
  }
  for (i = 0; i < MAX_PLAYER; i++) {
    int p = (current_dealer() + i) % 4;
    pool_draw(p);
  }

  wall_bones = make_wall_bones(pool_tomb);
  open_prize_marker();
}

int pool_draw(int pid)
{
  int n = *pool_head++;
  players[pid].hand[n]++;
  players[pid].dead_tiles[n]++;
  return n;
}

int pool_draw_after_quadruplet(int pid)
{
  int n = *--pool_tail;
  pool_tomb--;
  players[pid].hand[n]++;
  players[pid].dead_tiles[n]++;

  /* פ */
  wall_bones = wall_bones->next;
  open_prize_marker();

  return n;
}

static int prize_of(int n)
{
  if (is_dragon(n)) {
    n++;
    if (n > 29) n = 27;
  }else if (is_wind(n)) {
    n++;
    if (n > 33) n = 30;
  } else {
    n++;
    if (number_of(n) == 0) n -= 9;
  }
  return n;
}

int count_prizes(int *hand, int ready)
{
  int number_of_markers = pool_count_quadruplets() + 1;
  int *marker = prize_marking;
  int i, count = 0;

  for (i = 0; i < number_of_markers; i++, marker -= 2) {
    count += hand[prize_of(*marker)];
  }
  if (ready) {
    /* ΢ɥ⸫ */
    marker = prize_marking + 1;
    for (i = 0; i < number_of_markers; i++, marker -= 2) {
      count += hand[prize_of(*marker)];
    }
  }
  return count;
}

int pool_discard(int pid, int n, int angled)
{
  int i;
  for (i = 0; i < MAX_PLAYER; i++) {
    if (i != pid)
      players[i].dead_tiles[n]++;
  }
  players[pid].hand[n]--;

  /* ΤפξϿ */
  players[pid].discard[n] = bucket_append(pid, n, angled);

  return n;
}

int pool_expose(int pid, int n)
{
  int i;

  if (players[pid].hand[n] < 1) {
    fprintf(stderr, "pool_expose(): Illegal exposure\n");

    print_hand(pid);
    printf(" %s\n", tile_face[n]);

    exit(1);
  }

  for (i = 0; i < MAX_PLAYER; i++) {
    if (i != pid)
      players[i].dead_tiles[n]++;
  }
  players[pid].hand[n]--;
  return n;
}

void pool_expose_robbing(int pid, int n, int type, int discarder)
{
  player_t *p = &players[pid];
  set_t set;
  switch (type) {
  case HOOK_TRIPLET:
    pool_expose(pid, n);
    pool_expose(pid, n);
    set = mark_add((4 - (pid - discarder)) % 4, make_set(n, TRIPLET));
    break;
  case HOOK_QUADRUPLET:
    pool_expose(pid, n);
    pool_expose(pid, n);
    pool_expose(pid, n);
    set = mark_add((4 - (pid - discarder)) % 4, make_set(n, QUADRUPLET));
    break;
  case HOOK_BELOW:
    pool_expose(pid, n + 1);
    pool_expose(pid, n + 2);
    set = mark_add(1, make_set(n, SEQUENCE));
    break;
  case HOOK_ABOVE:
    pool_expose(pid, n - 2);
    pool_expose(pid, n - 1);
    set = mark_add(3, make_set(n - 2, SEQUENCE));
    break;
  case HOOK_MIDDLE:
    pool_expose(pid, n - 1);
    pool_expose(pid, n + 1);
    set = mark_add(2, make_set(n - 1, SEQUENCE));
    break;
  default:
    fprintf(stderr, "pool_expose_robbing(): Illegal type.\n");
    exit(1);
  }
  p->revealed[p->revealed_count++] = set;
  p->condition |= bit(COND_REVEALED); /* Ϫե饰 */

  /* ĥפξ򵭲 */
  players[discarder].discard[n] = exposure_append(pid, set);
  /* Τפ */
  players[discarder].bucket = players[discarder].bucket->next;  
}

/* ⤫ʻҤϪ롣ʤξ硢1֤ */
int pool_expose_quadruplet(int pid, int n)
{
  player_t *p = &players[pid];
  int i;
  switch (p->hand[n]) {
  case 4:
    pool_expose(pid, n);
    pool_expose(pid, n);
    pool_expose(pid, n);
    pool_expose(pid, n);
    p->revealed[p->revealed_count++] = toggle_revealed_bit(make_set(n, QUADRUPLET));

    exposure_append_quadruplet(pid, n);
    return 0;
  case 1:
    pool_expose(pid, n);
    for (i = 0; i < p->revealed_count; i++) {
      set_t set = p->revealed[i];
      if (tile_of(set) == n && type_of(set) == TRIPLET) {
        p->revealed[i] = mark_add(mark_of(set), make_set(n, QUADRUPLET));

        /* פξ򵭲 */
        players[pid].discard[n] = exposure_append_quadruplet_lifting(pid, n);
        return 1;
      }
    }
    /* ȴ */
  default:
    fprintf(stderr, "pool_expose_quadruplet(): Illegal quadruplet.\n");
    exit(1);
  }
  return 0;
}
