/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: round.c,v 1.3 1999/11/10 04:54:37 shinmaru Exp $
 */

#include "mahjong.h"
#include "hand.h"
#include "option.h"
#include "debug.h"
#include "stub.h"
#include <assert.h>

#ifdef __WIN32__
__declspec(dllimport) void __stdcall Sleep(unsigned);
#define usleep(s)	Sleep((s) / 1024)
#else
#include <unistd.h>
#endif

/* ξԤΰ㤤Ͼʶ路 */
#define is_final()	(bittest(__status, STATUS_FINAL))
#define is_finished()	(pool_finished())

#define delay()		{ if (pid != 0 && option_wait) usleep(400000); }

#define permission_of_sequences() (bit(HOOK_BELOW) | bit(HOOK_ABOVE) | bit(HOOK_MIDDLE))
#define permission_of_triplets() \
  (bit(HOOK_TRIPLET) | (pool_count_quadruplets() == 4 ? 0 : bit(HOOK_QUADRUPLET)))

player_t players[MAX_PLAYER];
int player_score[MAX_PLAYER];
int __discarder;
roundkeeper_t roundkeeper;
char round_caption[32];

static event_t __continuation;
static int __status;
static int __winds[4];

static event_t perform_dealing(void);
static event_t perform_drawing(int pid, int n);
static event_t perform_discard(int pid, int n);
static event_t perform_passing(int pid, int n);
static event_t perform_contention(int pid, int n);
static event_t perform_round_over(void);
static event_t loop_winning_report(void);
static event_t report_winnings(int pid, int n);
static event_t report_round_over(void);
static event_t report_ending_round(void);
static event_t force_discard(int pid);
static event_t expose_robbing(int pid, int n, int types);
static event_t expose_quadruplet(int pid, int n);

typedef struct loop_event_node {
  event_t event;
  struct loop_event_node *next;
} loop_event_t;

static loop_event_t *__notice;
static loop_event_t *__report;

#define MAX_LOOP		32

static loop_event_t *push_event(int event, loop_event_t *next)
{
  static loop_event_t buffer[MAX_LOOP], *end;
  loop_event_t *top;
  
  if (end == NULL) {
    int marked[MAX_LOOP];
    int i;
    loop_event_t *c;

    for (i = 0; i < MAX_LOOP; i++) {
      marked[i] = 0;
    }
    for (c = next; c != NULL; c = c->next)
      marked[c - buffer] = 1;
    for (c = __notice; c != NULL; c = c->next) {
      marked[c - buffer] = 1;
    }
    for (c = __report; c != NULL; c = c->next) {
      marked[c - buffer] = 1;
    }
    for (i = 0; i < MAX_LOOP; i++) {
      if (!marked[i])
        buffer[i].next = end, end = &buffer[i];
    }
    if (end == NULL) {
      fprintf(stderr, "event_push(): Out of memory.\n");
      exit(1);
    }
  }
  top = end; end = end->next;
  top->event = event;
  top->next = next;
  return top;
}

/* ΤΥ롼פ˿ */
static int loop_notification(int pid, int n, int permission, event_t c)
{
  int triplet_bits = permission_of_triplets();
  int i;
  assert(pid == __discarder);
  
  __continuation = c;
  for (i = next_seat(pid); i != pid; i = next_seat(i)) {
    __notice = push_event(event_notify_discard(i, n, permission), __notice);
    permission &= triplet_bits;
  }
  return REPLY_NEXT;
}

/* ϢǤΥå */
static void check_4_winds(int n)
{
  if (is_wind(n)) {
    if (++__winds[n - WIND_EAST] == 4)
      __status |= bit(STATUS_4_WINDS);
  }
}

/* ʻλΥå */
static void check_4_quadruplets(int pid)
{
  player_t *p = &players[pid];

  if (pool_count_quadruplets() == 4) {
    int i;
    if (p->revealed_count < 4) {
      __status |= bit(STATUS_4_QUADRUPLETS);
      return;
    }
    for (i = 0; i < p->revealed_count; i++) {
      if (type_of(p->revealed[i]) != QUADRUPLET) {
        __status |= bit(STATUS_4_QUADRUPLETS);
        return;
      }
    }
  }
}

/* ͲΩľΥå */
static void check_4_people_ready(void)
{
  int i;
  for (i = 0; i < MAX_PLAYER; i++) {
    if (!bittest(players[i].condition, COND_READY))
      return;
  }
  __status |= bit(STATUS_4_PEOPLE_READY);
}

/* 䡦ȯӥåȤ򥯥ꥢ */
static void reset_all_immediate(void)
{
  int i;
  for (i = 0; i < MAX_PLAYER; i++) {
    players[i].condition &= ~(bit(COND_FIRST_TURN) | bit(COND_IMMEDIATE));
  }
}

/* Τפθƨ */
static void through_discard(int discarder, int n)
{
  int i;
  for (i = 0; i < MAX_PLAYER; i++) {
    if (i != discarder)
      players[i].through[n] = players[discarder].discard[n];
  }
}

/* Ĵ٤ */
int tell_completion_bonus(evaluated_hand_t *e)
{
  int condition = players[e->pid].condition;
  int completions = 0;
  int self_drawn = e->pid == __discarder;

  /* ŷ¡ */
  if (self_drawn && bittest(condition, COND_FIRST_TURN)) {
    if (e->pid == current_dealer())
      completions |= bit(HEAVENLY_HAND); /*  */
    else
      completions |= bit(EARTHLY_HAND); /*  */
  }
  /*  */
  if (self_drawn && !e->revealed) {
    completions |= bit(TOTALLY_CONCEALED);
  }
  /* Ωľ֥Ωľ */
  if (bittest(condition, COND_READY)) {
    if (bittest(condition, COND_READY_DOUBLED))
      completions |= bit(DECLARED_READY_DOUBLED);
    else
      completions |= bit(DECLARED_READY);
  }
  /* ȯ */
  if (bittest(condition, COND_IMMEDIATE)) {
    completions |= bit(WINNING_IMMEDIATELY);
  }
  /* η */
  if (bittest(__status, STATUS_FINAL)) {
    if (self_drawn)
      completions |= bit(UNDER_THE_SEA);
    else
      completions |= bit(UNDER_THE_RIVER);
  }
  /* 峫֡ */
  if (bittest(__status, STATUS_AFTER_QUADRUPLET)) {
    if (self_drawn)
      completions |= bit(WINNING_AFTER_A_QUADRUPLET);
    else
      completions |= bit(ROBBING_A_QUADRUPLET);
  }
  return completions;
}

static void set_round_caption(void)
{
  roundkeeper_t *k = &roundkeeper;
  sprintf(round_caption, k->sticks == 0 ? "%s%d" : "%s%d%dܾ",
	  tile_face[k->round_wind + WIND_EAST], k->rounds + 1, k->sticks);
}

void start_round(int first_dealer)
{
  int i;

  for (i = 0; i < MAX_PLAYER; i++) {
    player_score[i] = INITIAL_FUNDAGE;
  }

  memset(&roundkeeper, 0, sizeof(roundkeeper));
  roundkeeper.first_dealer = first_dealer;

  set_round_caption();

  dispatch(ACTION_DEAL);
}

void dispatch(event_t e)
{
  for (;;) {
    switch (action_of(e)) {
    case ACTION_DEAL:
      e = perform_dealing();
      break;
      
    case ACTION_DRAW:
      {
        int pid = pid_of(e);
        int n = pool_draw(pid);
        if (pool_finished()) {
          __status |= bit(STATUS_FINAL);
        }
        __discarder = pid;
        e = perform_drawing(pid, n);
      }
      break;

    case ACTION_READY:
      {
        int pid = pid_of(e);
        int n = tile_of(e);

        players[pid].condition |= bit(COND_ONLY_DECLARED)
          | bit(COND_NOT_YET_INDICATED)
	  | bit(COND_IMMEDIATE);
        
        __status &= ~bit(STATUS_AFTER_QUADRUPLET);
        __discarder = pid;
        e = perform_discard(pid, n);
      }
      break;
      
    case ACTION_DISCARD:
      {
        int pid = pid_of(e);
        int n = tile_of(e);

	players[pid].condition &= ~bit(COND_IMMEDIATE);

        __status &= ~bit(STATUS_AFTER_QUADRUPLET);
        __discarder = pid;
        e = perform_discard(pid, n);
      }
      break;

    case ACTION_PASS:
      e = perform_passing(pid_of(e), tile_of(e));
      break;

    case ACTION_WIN:
      {
	int pid = pid_of(e);
	int n = tile_of(e);

	open_hand(pid);
	open_wall(); /* ΢ɥ򳫤 */

	e = report_winnings(pid, n);
      }
      break;

    case ACTION_ROB:
      reset_all_immediate();
      e = expose_robbing(pid_of(e), tile_of(e), permission_of(e));
      break;

    case ACTION_CONTENTION:
      e = perform_contention(pid_of(e), tile_of(e));
      break;
      
    case ACTION_QUADRUPLE:
      reset_all_immediate();
      __status |= bit(STATUS_AFTER_QUADRUPLET);
      e = expose_quadruplet(pid_of(e), tile_of(e));
      break;

    case ACTION_AFTER_QUADRUPLET_LIFTING:
      /* Ť֤ĥפξ̤פȤƵϿ */
      through_discard(pid_of(e), tile_of(e));

      /* ȴ */

    case ACTION_AFTER_QUADRUPLET:
      {
        int pid = pid_of(e);
        int n = pool_draw_after_quadruplet(pid_of(e));

        check_4_quadruplets(pid); /* 弫ΤƤĴ٤ */
	repaint_wall(); /* פι */

        __status |= bit(STATUS_AFTER_QUADRUPLET);
        __discarder = pid;
        e = perform_drawing(pid, n);
      }
      break;
      
    case ACTION_ROUND_OVER:
      e = perform_round_over();
      break;

      /* Υ롼 */
    case NOTIFY_DISCARD:
      {
        int pid = pid_of(e);
        int n = tile_of(e);
        int permission = permission_of(e);

        e = option_batch || pid != 0 ? batch_notified(pid, n, permission)
          : interactive_notified(pid, n, permission);
      }
      break;
      
    case REPLY_WIN:
      {
	int pid = pid_of(e);
	int n = tile_of(e);
	__report = push_event(event_report_winning(pid, n), __report);
      }
      /* ȴ */
      
    case REPLY_NEXT:
      if (__notice == NULL) {
        if (__report != NULL)
	  e = loop_winning_report();
        else
          e = __continuation;
      } else {
        e = __notice->event, __notice = __notice->next;
      }
      break;

      /* 롼 */
    case REPORT_WINNING:
      e = report_winnings(pid_of(e), tile_of(e));
      break;

    case REPORT_OK:
      if (__report == NULL) {
        e = ACTION_ROUND_OVER;
      } else {
        e = __report->event, __report = __report->next;
      }
      break;

    default:
      return;
    }
  }
}

/*  */
event_t perform_dealing(void)
{
  int i;
  memset(players, 0, sizeof(players));
  memset(__winds, 0, sizeof(__winds));
  pool_shuffle();
  pool_deal();
  for (i = 0; i < MAX_PLAYER; i++) {
    hook_table_t h;
    
    /* פɾ */
    players[i].going = calculate_hand(players[i].hand, h);
    if (players[i].going == 1) {
      memcpy(players[i].winnings, h, sizeof(h));
    }
    players[i].condition = bit(COND_FIRST_TURN);
    align_handrail(i, -1, SURFACE_HEAD); /*  */
  }

  __status = 0;
  __report = NULL;
  __notice = NULL;

  repaint_all();
  return event_drawing(current_dealer());
}

/*  */
event_t perform_drawing(int pid, int n)
{
  int forbids = is_final() || pool_count_quadruplets() == 4;

  align_handrail(pid, n, SURFACE_HEAD); /*  */
  repaint_handrail(pid);

  return option_batch || pid != 0 ? batch_choose(pid, n, forbids)
    : interactive_choose(pid, n, forbids);
}

/* ׽ */
event_t perform_discard(int pid, int n)
{
  player_t *p = &players[pid];
  int angled;
  event_t c;
  int permission;
  int i;
 
  if (bittest(p->condition, COND_FIRST_TURN)) {
    p->condition &= ~bit(COND_FIRST_TURN);

    check_4_winds(n);
    if (bittest(p->condition, COND_ONLY_DECLARED)) {
      /* ֥ΩľΩľ򽤾
       * (ΩľΤϤޤꤷʤ) */
      p->condition |= bit(COND_READY_DOUBLED);
    }
  }
  else if (!is_ready(pid)) {
    for (i = 0; i < MAX_SORT; i++) 
      players[pid].through[i] = NULL;
  }

  if (pid != 0 && option_wait) {
    bone_t *b;
    for (b = players[pid].handrail.bones; b != NULL; b = b->next) {
      if (n == tile_of(b->tid)) {
	b->tid = toggle_disappear(b->tid);
	break;
      }
    }
    delay();
    if (b != players[pid].handrail.bones) {
      /* Ф */
      repaint_handrail(pid);
    }
    usleep(100000);
  }

  angled = bittest(p->condition, COND_NOT_YET_INDICATED);  
  pool_discard(pid, n, angled);

  align_handrail(pid, -1, SURFACE_HEAD); /*  */
  repaint_bucket(pid);
  repaint_handrail(pid);

#ifdef TRACE
  printf("%s %s\n", tile_face[seat_wind_of(pid)], tile_face[n]);
#endif

  if (is_finished()) {
    c = event_pass(pid, n);
    permission = 0;
  } else {
    c = event_contention(pid, n);
    permission = permission_of_sequences() | permission_of_triplets();
  }

  return loop_notification(pid, n, permission, c);
}

event_t perform_contention(int pid, int n)
{
  int bits_triplet = permission_of_triplets();
  int permission =  bits_triplet | permission_of_sequences();
  int robber = -1;
  int robbing_types = 0;
  int i;

  /* 󥿥饯ƥ֤ʥץ쥤䡼ξ硢loop_notification()ؿ
   * 륤٥ȽĤμͽᤵ롣
   * λǤϡɤΥץ쥤䡼Ĥͥ褹뤫ϰդ˷ޤ */
  assert(pid == __discarder);

  if (bittest(players[pid].condition, COND_ONLY_DECLARED)) {
    /* ̤Ωľꡣ
     * (פξ硢ˤʤΩľˤϰ̣ʤ) */
    players[pid].condition |= bit(COND_READY);
    players[pid].condition &= ~bit(COND_ONLY_DECLARED);

    player_score[pid] -= READY_EXPENSE;
    roundkeeper.carried_over[pid]++;

    check_4_people_ready();
  }

  for (i = next_seat(pid); i != pid; i = next_seat(i)) {
    if (!is_ready(i)) {
      int types = players[i].robbings[n].types & permission;
      if (types != 0) {
        robber = i;
        robbing_types = types;
      }
      permission = bits_triplet;
    }
  }
  return robber >= 0 ? event_robbing(robber, n, robbing_types)
    : perform_passing(pid, n);
}

/* οͤ˼֤ */
event_t perform_passing(int pid, int n)
{
  assert(pid == __discarder);
  
  if (bittest(__status, STATUS_4_QUADRUPLETS)) {
    return interactive_report_round_over("ʻλ", NULL);
  }
  if (bittest(__status, STATUS_4_WINDS)) {
    return interactive_report_round_over("Ϣ", NULL);
  }
  if (bittest(__status, STATUS_4_PEOPLE_READY)) {
    return interactive_report_round_over("ͲΩľ", NULL);
  }
  if (pool_finished()) {
    return report_round_over();
  }

  players[pid].condition &= ~bit(COND_NOT_YET_INDICATED);
  through_discard(pid, n);

#ifdef TRACE
  printf("%s \n", tile_face[seat_wind_of(next_seat(pid))]);
#endif

  return event_drawing(next_seat(pid));
}

static event_t loop_winning_report(void)
{
  loop_event_t *e;
  int count = 0;

  for (e = __report; e != NULL; e = e->next) {
    int pid = pid_of(e->event);

    if (players[pid].scoresheet != NULL) {
      count++;
    }
    open_hand(pid);
  }

  /* ȱ¤Υå */
  if (count < 3) {
    open_wall(); /* ΢ɥ򳫤 */
    return REPORT_OK;
  } else {
    /* 򥭥󥻥뤷ήɤ */
    __report = NULL;
    __status |= bit(STATUS_CONTINUE_DEALER); /* Ϣ */
    return interactive_report_round_over("ȱ", NULL);
  }
}

/* ο */
event_t report_winnings(int pid, int n)
{
  int bill[MAX_PLAYER];  
  char reason[256];
  int i;
  event_t e;

  if (players[pid].scoresheet == NULL ||
      pid == current_dealer()) {
    __status |= bit(STATUS_CONTINUE_DEALER); /* Ϣ */
  }

  if (players[pid].scoresheet == NULL) {
    sprintf(reason, "%s ", tile_face[seat_wind_of(pid)]);
    fined_penalty(pid, bill);
    e = interactive_report_winning(reason, pid, n, bill);
  }
  else if (pid == __discarder) {
    sprintf(reason, "%s ", tile_face[seat_wind_of(pid)]);    
    collect_money_each(pid, bill);
    e = interactive_report_winning(reason, pid, n, bill);
  }
  else {
    /* ΩľƼԤβȤ */
    sprintf(reason, "%s ", tile_face[seat_wind_of(pid)]);
    collect_money_loser(pid, __discarder, bill);
    e = interactive_report_winning(reason, pid, n, bill);
  }

  for (i = 0; i < MAX_PLAYER; i++) {
    player_score[i] += bill[i];
  }
  return e;
}

/* İפο */
event_t report_round_over(void)
{
  int ready[MAX_PLAYER], count = 0;
  int bill[MAX_PLAYER];
  int i;
  
  for (i = 0; i < MAX_PLAYER; i++) {
    hook_table_t h;

    ready[i] = calculate_hand(players[i].hand, h) == 1;
    if (ready[i]) {
      count++;
      open_hand(i);
    }
  }
  if (count == 4 || count == 0) {
    for (i = 0; i < MAX_PLAYER; i++)
      bill[i] = 0;
  } else {
    for (i = 0; i < MAX_PLAYER; i++)
      bill[i] = ready[i] ? (3000 / count) : -(3000 / (4 - count));
  }

  if (ready[current_dealer()]) {
    /* ƤİפʤϢ */
    __status |= bit(STATUS_CONTINUE_DEALER); 
  } else {
    __status |= bit(STATUS_CARRY_STICKS);
  }

  for (i = 0; i < MAX_PLAYER; i++) {
    if (bill[i] != 0)
      return interactive_report_round_over("ή", bill);
  }
  return interactive_report_round_over("ή", NULL);
}

event_t perform_round_over(void)
{
  roundkeeper_t *k = &roundkeeper;

  if (dealer_continued(__status)) {
    k->sticks++;
  } else {
    if (bittest(__status, STATUS_CARRY_STICKS)) {
      k->sticks++;
    } else {
      k->sticks = 0;
    }
    k->rounds = (k->rounds + 1) % 4;
    if (k->rounds == 0) {
      if (++k->round_wind >= 2)
	return report_ending_round();
    }
  }
  set_round_caption();

  return ACTION_DEAL;
}

event_t report_ending_round(void)
{
  int ranking[MAX_PLAYER], bill[MAX_PLAYER];
  int point, count;
  int i, j, rank;
  static const int ranking_point[MAX_PLAYER] = { 12000, 4000, -4000, -12000 };

  count = 0;
  for (i = 0; i < MAX_PLAYER; i++) {
    int s = player_score[i];
    rank = 0;
    for (j = 0; j < MAX_PLAYER; j++) {
      if (player_score[j] > s)
	rank++;
    }
    ranking[i] = rank;
    if (rank == 0) count++;
  }

  /* ȥå׾ޤʬ */
  point = (30000 - INITIAL_FUNDAGE) * 4 / count;

  for (i = 0; i < MAX_PLAYER; i++) {
    bill[i] = ranking[i] == 0 ? point : 0;
    player_score[i] -= 30000;
  }

  if (count == 1) {
    int carried_over = 0;
    int top;
    /* ۤ줿Ωľϥȥåפ㤦 */
    for (i = 0; i < MAX_PLAYER; i++) {
      carried_over += roundkeeper.carried_over[i];
      if (ranking[i] == 0)
	top = i;
    }
    bill[top] += carried_over * READY_EXPENSE;
  } else {
    /* ۤ줿Ωľϻʧ֤롣
     * ΤȤ̤ưʤ */
    for (i = 0; i < MAX_PLAYER; i++)
      bill[i] += roundkeeper.carried_over[i] * READY_EXPENSE;
  }

  /* ʬ */
  for (rank = 0; rank < MAX_PLAYER; rank++) {
    point = 0;
    count = 0;
    for (i = 0; i < MAX_PLAYER; i++) {
      if (ranking[i] == rank)
	point += ranking_point[rank + count++];
    }
    if (count > 0) {
      point /= count;
      for (i = 0; i < MAX_PLAYER; i++) {
	if (ranking[i] == rank)
	  bill[i] += point;
      }
    }
  }

  sprintf(round_caption, "Ⱦλ");
  interactive_report_round_over(NULL, bill);
  return ACTION_SUSPEND;
}

event_t force_discard(int pid)
{
  const int forbidden = 1;

  return option_batch || pid != 0 ? batch_choose(pid, -1, forbidden)
    : interactive_choose(pid, -1, forbidden);
}

/* Ϫη򤹤 */
static int select_robbing_type(int types)
{
  if ((types & permission_of_sequences()) != 0) {
    if (bittest(types, HOOK_MIDDLE))
      return HOOK_MIDDLE;
    else
      return bittest(types, HOOK_BELOW) ? HOOK_BELOW : HOOK_ABOVE;
  } else
    return bittest(types, HOOK_TRIPLET) ? HOOK_TRIPLET : HOOK_QUADRUPLET;
}

/* ̻ҤϪ */
event_t expose_robbing(int pid, int n, int permission)
{
  int type = select_robbing_type(permission);

  /* ʹߡϪե饰ʤ̯ʥХͶȯ롣
   * 㤨Сݥ󤷤¾ïݥ󥫥ΤƤȤ
   * βϴְäݥ󤷤褦Ȥǽ롣
   * μ¤ʤΤϡǤ٤ƤϪե饰ꥻåȤ뤳ȤǤ */
#if 0
  memset(players[pid].robbings, 0, sizeof(hook_table_t));
#endif

  pool_expose_robbing(pid, n, type, __discarder);
  through_discard(__discarder, n); /* ̤פξĥפξǾ */

  align_handrail(pid, -1, SURFACE_HEAD); /*  */
  repaint_handrail(pid);
  repaint_bucket(__discarder);

  delay();

  return type == HOOK_QUADRUPLET ? event_after_quadruplet(pid)
    : force_discard(pid);
}

/* ʻҤϪ */
event_t expose_quadruplet(int pid, int n)
{
  int lifted = pool_expose_quadruplet(pid, n);

#ifdef TRACE
  printf("%s  %s\n", tile_face[seat_wind_of(pid)], tile_face[n]);
#endif

  align_handrail(pid, -1, SURFACE_HEAD); /*  */
  repaint_handrail(pid);

  delay();

  if (lifted)
    /* ʤα¤Ĵ٤ */
    return loop_notification(pid, n, 0, event_after_quadruplet_lifting(pid, n));
  else
    return event_after_quadruplet(pid);
}
