/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: score.c,v 1.2 1999/11/06 01:07:50 shinmaru Exp $
 */

#include "mahjong.h"
#include "hand.h"

#define ceiling_100(n)		((((n) + 99) / 100) * 100)

#define MAX_HAND		27
#define MAX_COMPLETION		11

#define TABLE_SIZE		(MAX_HAND + MAX_COMPLETION)

#define GRADE_1			1	/* ӵ */
#define GRADE_2			2	/* ķ */
#define GRADE_3			3	/*  */
#define GRADE_4			4	/*  */
#define LIMITED			5	/*  */

static const char *__fmt[] = {
  "%d%d",
  "",
  "ķ",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "ϻ", /* ʹ()쿧ͰŹñ()ʻҤȤ */
  "" /* ʼ */
};

static int __fmt_last = sizeof(__fmt)/sizeof(__fmt[0]) - 1;

static const int limited_score[6][2] = {
  {     0,     0 },
  {  8000, 12000 },
  { 12000, 18000 },
  { 16000, 24000 },
  { 24000, 36000 },
  { 32000, 48000 }
};

static const struct {
  char *name;
  int bonus[2];
  int grade;
} score_table[TABLE_SIZE] = {
  { "",		{  1,  1 },       0 },
  { "ʿ",		{  1,  0 },       0 },
  { "쵤̴",		{  2,  1 },       0 },
  { "ָ",		{  1,  0 },       0 },
  { "ָ",		{  3,  0 },       0 },
  { "",		{  2,  1 },       0 },
  { "",	{  3,  2 },       0 },
  { "ϷƬ",		{  2,  2 },       0 },
  { "ϷƬ",		{  1,  1 }, LIMITED },
  { "Ʊ",		{  2,  1 },       0 },
  { "Ʊ",		{  2,  2 },       0 },
  { "Ź",		{  2,  2 },       0 },
  { "ͰŹ",		{  1,  1 }, LIMITED },
  { "ʻ",		{  2,  2 },       0 },
  { "ʻ",		{  1,  1 }, LIMITED },
  { "С",		{  2,  2 },       0 },
  { "л",		{  2,  0 },       0 },
  { "쿧",		{  3,  2 },       0 },
  { "쿧",		{  6,  5 },       0 },
  { "а쿧",		{  1,  1 }, LIMITED },
  { "쿧",		{  1,  1 }, LIMITED },
  { "",		{  2,  2 },       0 },
  { "绰",		{  1,  1 }, LIMITED },
  { "ʹ",		{  1,  1 }, LIMITED },
  { "ʹ",		{  1,  1 }, LIMITED },
  { "ϡ",		{  1,  1 }, LIMITED },
  { "̵",		{  1,  1 }, LIMITED },
  /*  */
  { "",	{  1,  0 },       0 },
  { "Ωľ",		{  1,  0 },       0 },
  { "֥Ωľ",	{  2,  0 },       0 },
  { "ȯ",		{  1,  0 },       0 },
  { "η",		{  1,  1 },       0 },
  { "",		{  1,  1 },       0 },
  { "",		{  1,  1 },       0 },
  { "峫",		{  1,  1 },       0 },
  { "",		{  1,  1 }, LIMITED },
  { "ŷ",		{  1,  1 }, LIMITED }
};

static int is_all_green(evaluated_hand_t *e);
static int is_nine_lanterns(int *hand);

#define BUFFER_SIZE		128

static hand_name_t *chain_name(hand_name_t *next)
{
  static hand_name_t buffer[BUFFER_SIZE], *end = NULL;
  hand_name_t *top;

  if (end == NULL) {
    int marked[BUFFER_SIZE];
    int i;
    hand_name_t *p;

    memset(marked, 0, sizeof(marked));
    for (p = next; p != NULL; p = p->next) {
      marked[p - buffer] = 1;
    }
    for (i = 0; i < MAX_PLAYER; i++) {
      if (players[i].scoresheet != NULL)
        for (p = players[i].scoresheet->names; p != NULL; p = p->next) {
          marked[p - buffer] = 1;
        }
    }
    for (i = 0; i < BUFFER_SIZE; i++) {
      if (!marked[i]) {
        buffer[i].next = end;
        end = &buffer[i];
      }
    }
    if (end == NULL) {
      fprintf(stderr, "chain_name(): Out of memory.\n");
      exit(1);
    }
  }
  top = end; end = end->next;
  top->next = next;
  top->bonus = 0;
  top->name = "";
  return top;
}

static hand_name_t *
list_name(int hands, int completions, int honors, int prizes, int revealed)
{
  hand_name_t *top = NULL;
  int i;

  if (hands == 0 && completions == 0 && honors == 0)
    return NULL; /* ʤ */

  revealed = !!revealed; /* 01ˤ */

  if (prizes > 0) {
    top = chain_name(top);
    top->name = "ɥ"; top->bonus = prizes;
  }
  if (honors > 0) {
    top = chain_name(top);
    top->name = ""; top->bonus = honors;
  }
  for (i = 0; i < MAX_HAND; i++) {
    if (bittest(hands, i)) {
      top = chain_name(top);
      top->name = score_table[i].name;
      if (score_table[i].grade > 0)
        top->bonus = 0;
      else
        top->bonus = score_table[i].bonus[revealed];
    }
  }
  for (i = MAX_COMPLETION - 1; i >= 0; i--) {
    if (bittest(completions, i)) {
      int j = i + MAX_HAND;
      top = chain_name(top);
      top->name = score_table[j].name;
      if (score_table[j].grade > 0)
        top->bonus = 0;
      else
        top->bonus = score_table[j].bonus[revealed];
    }
  }
  return top;
}

static int graded_hands[5], graded_completions[5];
static int gathered;

static void gather_wonderful_hands(void)
{
  int i;
  for (i = 0; i < MAX_HAND; i++) {
    int grade = score_table[i].grade;
    if (grade > 0)
      graded_hands[grade] |= bit(i);
  }
  for (i = 0; i < MAX_COMPLETION; i++) {
    int grade = score_table[i + MAX_HAND].grade;
    if (grade > 0)
      graded_completions[grade] |= bit(i);
  }
}

scoresheet_t *
compute_score(int pid, int hands, int completions, int honors, int prizes, int point,
              int revealed)
{
  static scoresheet_t sheet;
  int is_dealer = pid == current_dealer();
  int grade, bonus = 0;
  int i;

  if (!gathered) {
    gather_wonderful_hands();
  }

  revealed = !!revealed; /* 01ˤ */
  is_dealer = !!is_dealer;

  for (grade = LIMITED; grade > 0; grade--) {
    if ((hands & graded_hands[grade]) != 0 ||
        (completions & graded_completions[grade]) != 0) {
      hands &= graded_hands[grade];
      completions &= graded_completions[grade];
      honors = 0;
      prizes = 0;
      break;
    }
  }
  if (grade == 0 || grade == LIMITED) {
    for (i = 0; i < MAX_HAND; i++) {
      if (bittest(hands, i))
        bonus += score_table[i].bonus[revealed];
    }
    for (i = 0; i < MAX_COMPLETION; i++) {
      if (bittest(completions, i))
        bonus += score_table[i + MAX_HAND].bonus[revealed];
    }
  }

  sheet.names = list_name(hands, completions, honors, prizes, revealed);
  if (sheet.names == NULL)
    return NULL;
  
  if (grade == 0) {
    int total;
    bonus += honors + prizes;
    switch (bonus) {
    case  0:
    case  1:
    case  2:
    case  3:
    case  4:
      if (point == 20 && bonus == 1)
        point = 30;
      total = (is_dealer ? 6 : 4) * point << (bonus + 2);
      total = (total + 99) / 100 * 100;
      if (total < limited_score[1][is_dealer]) break;
      /* ȴ */
    case  5: grade = GRADE_1; break;
    case  6:
    case  7: grade = GRADE_2; break; 
    case  8: 
    case  9: 
    case 10: grade = GRADE_3; break;
    case 11:
    case 12: grade = GRADE_4; break; 
    default: grade = LIMITED; break;
    }
    if (grade > 0) {
      total = limited_score[grade][is_dealer];
    }
    sheet.point = point;
    sheet.bonus = bonus;
    sheet.total = total;
    sheet.fmt = __fmt[grade];
  } else {
    sheet.point = 0;
    sheet.bonus = 0;
    sheet.total = limited_score[grade][!!is_dealer];
    if (grade < LIMITED) {
      sheet.fmt = __fmt[grade];
    } else {
      sheet.total *= bonus;
      sheet.fmt = __fmt[max(__fmt_last, bonus + LIMITED - 1)];
    }
  }
  return &sheet;
}

#if 1
#define hand_append(m, h)	((m) |= bit(h))
#else
#define hand_append(m, h)	(printf("add: %s\n", score_table[h].name), (m) |= bit(h))
#endif

/* 1ϡּספǤϤʤϪ̻Ҥޤ᤿׼Υޥå */
int judge(int *hand, int *values, evaluated_hand_t *e)
{
  int mahjong = 0;
  int dragons = 0, winds = 0, concealed_triplets = 0, quadruplets = 0;
  int honors;
  int i;

  /* Ʊ */
  for (i = 0; i < 9; i++) {
    if (e->triplets[i] + e->triplets[i + 9] + e->triplets[i + 18] == 9) {
      mahjong |= bit(EQUAL_TRIPLETS_IN_THREE_SUITS);
      break;
    }
  }

  for (i = 0; i < e->count; i++) {
    set_t set = e->sets[i];
    switch (type_of(set)) {
    case QUADRUPLET: quadruplets++;
      /* ȴ */
    case TRIPLET:
      if (!is_revealed(set))
	concealed_triplets++;
      break;
    }
  }
  
  /* ŹͰŹ */
  if (concealed_triplets == 4)
    hand_append(mahjong, FOUR_CONCEALED_TRIPLETS); /*  */
  else if (concealed_triplets == 3)
    hand_append(mahjong, THREE_CONCEALED_TRIPLETS);

  /* ʻҡʻ */
  if (quadruplets == 4)
    hand_append(mahjong, FOUR_QUADRUPLETS); /*  */
  else if (quadruplets == 3)
    hand_append(mahjong, THREE_QUADRUPLETS);

  for (i = 27; i < 30; i++) dragons += hand[i];
  for (i = 30; i < 34; i++) winds += hand[i];
  honors = dragons + winds;

  /* 绰 */
  if (dragons == 9)
    hand_append(mahjong, MAJOR_THREE_DRAGONS); /*  */
  else if (dragons == 8)
    hand_append(mahjong, MINOR_THREE_DRAGONS);
  
  /*  */
  if (values[EVAL_NO_TERMINALS] == 14) {
    hand_append(mahjong, NO_TERMINALS);
  }
  /* 󥿡 */
  if (values[EVAL_TERMINALS] == 14) {
    if (honors == 0)
      hand_append(mahjong, ALL_TERMINALS_WITH_NO_HONORS);
    else
      hand_append(mahjong, ALL_TERMINALS);
  }
  /* ʿ */
  if (values[EVAL_SEQUENCES] == 14)
    hand_append(mahjong, ALL_SEQUENCES);
  /* ָָ */
  if (values[EVAL_SISTERS] == 12)
    hand_append(mahjong, DUAL_SISTER_SEQUENCES);
  else if (values[EVAL_SISTERS] >= 6)
    hand_append(mahjong, SISTER_SEQUENCES);
  /* 쵤̴ */
  if (values[EVAL_STRAIGHT] >= 9)
    hand_append(mahjong, STRAIGHT_SEQUENCES);
  /* Ʊ */
  if (values[EVAL_3_SUIT_EQUALS] >= 9)
    hand_append(mahjong, EQUAL_SEQUENCES_IN_THREE_SUITS);
  /* С¡ϷƬϷƬ */
  if (values[EVAL_TRIPLETS] == 14) {
    if (honors == 14)
      hand_append(mahjong, ALL_HONORS); /*  */
    else if (bittest(mahjong, ALL_TERMINALS)) {
      mahjong &= ~bit(ALL_TERMINALS);
      hand_append(mahjong, ALL_PRIMARIES);
    } else if (bittest(mahjong, ALL_TERMINALS_WITH_NO_HONORS))
      hand_append(mahjong, ALL_PRIMARIES_WITH_NO_HONORS); /*  */
    else
      hand_append(mahjong, ALL_TRIPLETS);
  }
  /* 쿧쿧а쿧Ϣ */
  if (values[EVAL_1_SUIT] == 14) {
    if (honors != 0) {
      if (is_all_green(e))
        hand_append(mahjong, ALL_GREENS); /*  */
      else
        hand_append(mahjong, ONE_SUIT);
    } else if (e->revealed && is_nine_lanterns(hand))
      hand_append(mahjong, NINE_LANTERNS); /*  */
    else
      hand_append(mahjong, ONE_SUIT_WITH_NO_HONORS);
  }
  /* л */
  if (values[EVAL_7_PAIRS] == 14)
    hand_append(mahjong, SEVEN_PAIRS);
  /* ̵ */
  if (values[EVAL_13_ORPHANS] == 14)
    hand_append(mahjong, THIRTEEN_ORPHANS); /*  */
  /* ʹʹ */
  if (winds == 12)
    hand_append(mahjong, MAJOR_FOUR_WINDS); /*  */
  else if (winds == 11)
    hand_append(mahjong, MINOR_FOUR_WINDS); /*  */

  return mahjong;
}

int is_nine_lanterns(int *hand)
{
  static const int nine_lanterns[9] = { 3, 1, 1, 1, 1, 1, 1, 1, 3 };
  int diff = 0;
  int i;
  
  for (i = 0; i < 9; i++) {
    int n = hand[i] - nine_lanterns[i];
    if (n < 0) return 0;
    if (n == 1)
      if (diff) return 0; else diff = 1;
  }
  return 1;
}

int is_all_green(evaluated_hand_t *e)
{
  int i;
  for (i = 0; i < e->count; i++) {
    set_t set = e->sets[i];
    int n = tile_of(set);
    switch (type_of(set)) {
    case SEQUENCE:
      if (n != 19)
        return 0;
    case PAIR:
    case TRIPLET:
    case QUADRUPLET:
      if (n != 19 || n != 20 || n != 21 || n != 23 || n != 25 || n != 28)
        return 0;
    default:
      return 0;
    }
  }
  return 1;
}

void collect_money_each(int pid, int *bill)
{
  int total = players[pid].scoresheet->total;
  int gratuity;
  int i;
  
  if (pid == current_dealer()) {
    int each = ceiling_100(total / 3);
    total = each * 3;
    for (i = 0; i < MAX_PLAYER; i++) {
      if (i != pid)
        bill[i] = -each;
    }
  } else {
    int half = ceiling_100(total / 2);
    int quarter = ceiling_100(half / 2);
    total = half + quarter * 2;
    for (i = 0; i < MAX_PLAYER; i++) {
      if (i != pid) {
        if (i == current_dealer())
          bill[i] = -half;
        else
          bill[i] = -quarter;
      }
    }
  }

  players[pid].scoresheet->total = total;

  bill[pid] = total;

  gratuity = roundkeeper.sticks * STICK_GRATUITY;
  for (i = 0; i < MAX_PLAYER; i++) {
    if (i != pid) {
      bill[i] -= gratuity;
      bill[pid] += gratuity;
    }
    bill[pid] += roundkeeper.carried_over[i] * READY_EXPENSE;
    roundkeeper.carried_over[i] = 0;
  }
}

void collect_money_loser(int pid, int loser, int *bill)
{
  int carried_over = 0;
  int i;
  for (i = 0; i < MAX_PLAYER; i++) {
    bill[i] = 0;
    carried_over += roundkeeper.carried_over[i];
    roundkeeper.carried_over[i] = 0;
  }

  bill[pid] += players[pid].scoresheet->total;
  bill[pid] += roundkeeper.sticks * STICK_GRATUITY * 3;

  bill[loser] = -bill[pid];

  bill[pid] += carried_over * READY_EXPENSE;
}

void fined_penalty(int pid, int *bill)
{
  int i;
  if (pid == current_dealer()) {
    for (i = 0; i < MAX_PLAYER; i++) {
      if (i != pid)
        bill[i] = 4000;
    }
    bill[pid] = -12000;
  } else {
    for (i = 0; i < MAX_PLAYER; i++) {
      if (i == current_dealer())
        bill[i] = 4000;
      else
        bill[i] = 2000;
    }
    bill[pid] = -8000;
  }
}
