/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: self_main.c,v 1.2 1999/11/06 01:07:51 shinmaru Exp $
 */

#include "mahjong.h"
#include "option.h"
#include "debug.h"
#include <math.h>
#include <time.h>

#define TURN_MAX	17

int self_play(int count);
int select_discard(int *);
int get_going(int *);

static int ready_turn_min = INT_MAX; /* û */
static double ready_average; /* ʿ */
static double ready_std; /* ɸк */
static int ready_total; /* İ׽ι */
static int ready_count; /* İײ */

static int winning_count; /* λ */
static int score_max; /*  */
static int score_total; /*  */

static int quadruple_in_hand(int pid, int n);
static int choose_one_suit(int *dead_tiles);

static void record_ready_turn(int turn)
{
  double x = turn - ready_average;

  ready_turn_min = min(ready_turn_min, turn);  
  ready_total += turn;
  ready_count++;
  ready_average += x / ready_count;
  ready_std += (ready_count - 1) * x * x / ready_count;
}

static void record_winning_count(int score)
{
  score_total += score;
  score_max = max(score_max, score);  
  winning_count++;
}

static void print_going(int going)
{
  if (going == 1)
    printf("   İ ");
  else
    printf("   %dİ", going - 1);
}

int main(int argc, char *argv[])
{
  int repeat = 0;
  int i;

  option_silence = 1;
  
  switch (argc) {
  case 1: repeat = 1; break;
  case 2: repeat = atoi(argv[1]); break;
  }
  if (repeat == 0) {
    fprintf(stderr, "self [repeat-count]\n");
    return 1;
  }

  roundkeeper.rounds = 3;

#if 1
  srandom((unsigned)time(NULL));
#else
  srandom(0x380d26b9);
#endif

  for (i = 0; i < repeat; i++) {
    self_play(i + 1);
    putc('.', stderr);
    putchar('\n');
  }

  printf("ײ: %d\n", repeat);
  printf("λ: %d\n", winning_count);
  printf("λΨ: %d\n", (int)(winning_count * 100.0 / repeat));
  if (winning_count != 0) {
    printf("    ǹ: %d\n", score_max);
    printf("    ʿ: %d\n", score_total / winning_count);
  }
  printf("İײ: %d\n", ready_count);  
  printf("İ׳Ψ: %d\n", (int)(ready_count * 100.0 / repeat));
  if (ready_count != 0) {
    printf("    û: %d\n", ready_turn_min);
    printf("    ʿ: %d\n", ready_total / ready_count);
    printf("ɸк: %f\n",
           (ready_count > 1 ? sqrt(ready_std / (ready_count - 1)) : 1.0));
  }
  return 0;
}

int self_play(int count)
{
  const int pid = 0;
  int turn, going;
  int i;
  hook_table_t h;

  memset(players, 0, sizeof(players));
  
  pool_shuffle();
  pool_deal();

  printf("%d\n", count);
  printf(" ");
  print_hand(pid);
  printf("              ");

  going = calculate_hand(players[pid].hand, h);
  print_going(going);
  if (going == 1) {
    print_hook_table(h);
  }
  putchar('\n');
  
  if (going == 1) {
    players[pid].condition |= bit(COND_READY) | bit(COND_READY_DOUBLED);
    record_ready_turn(0); /* İ׽ܤϿ */
  }

  for (turn = 1; !pool_finished(); turn++) {
    int ready_now;
    int going;
    int n1, n2;
    
    n1 = pool_draw(pid);
    
  evaluation:
    ready_now = 0;
    evaluate(pid, n1);
    
    if (players[pid].scoresheet != NULL) {
      printf("%2d!", turn);

      players[pid].hand[n1]--;
      print_hand(pid);
      printf(" :%s\n", tile_face[n1]);

      print_scoresheet(players[pid].scoresheet);
      record_winning_count(players[pid].scoresheet->total); /* λ˵Ͽ */
      break;
    } else {
      n2 = choose_one_suit(players[pid].dead_tiles);
      
      going = refer_going(n2);
      if (!is_ready(pid) && going == 1) {
        players[pid].condition |= bit(COND_READY);
        ready_now = 1;
        record_ready_turn(turn); /* İ׽ܤϿ */
      }
      if (going_table[n2].quadrupling) {
        /*  */
        printf(ready_now ? "%2d*" : "%2d'", turn);
        print_hand(pid);
        printf(" :%s \n", tile_face[n1]);

        n1 = quadruple_in_hand(pid, n2);
        goto evaluation;
      } else {
        printf(ready_now ? "%2d*" : "%2d ", turn);
        pool_discard(pid, n2, ready_now);

        print_hand(pid);
        printf(" :%s :%s", tile_face[n1], tile_face[n2]);
        print_going(going);

        if (going == 1 && n1 != n2) {
          print_hook_table(going_table[n2].hook);
        }
        putchar('\n');
      }
    }

    /* ¾λȤϼڤ */
    for (i = 1; i < MAX_PLAYER; i++) {
      int n = pool_draw(i);
      pool_discard(i, n, 0);
    }
  }
  return turn;
}

/* ׷ */
int choose_one_suit(int *dead_tiles)
{
  int max_gone = 0, max_weight = 0;
  int result;  
  int i;

  for (i = 0; i < MAX_SORT; i++) {
    if (max_gone < going_table[i].max_gone)
      max_gone = going_table[i].max_gone;
  }
  
  for (i = MAX_SORT - 1; i >= 0; i--) {
    int weight = 0;
    if (going_table[i].max_gone == max_gone) {
      int n;
      for (n = 0; n < MAX_SORT; n++) {
        int rest_count = 4 - dead_tiles[n];
        if (rest_count >= going_table[i].hook[n].needs) {
          weight += rest_count;
        }
        if (bittest(going_table[i].hook[n].types, HOOK_ABOVE))
          weight += 4;
        if (bittest(going_table[i].hook[n].types, HOOK_BELOW))
          weight += 4;
        if (bittest(going_table[i].hook[n].types, HOOK_MIDDLE))
          weight += 6;
      }
      weight += going_table[i].values[EVAL_1_SUIT];
      if (max_weight < weight) {
        max_weight = weight;
        result = i;
      }
    }
  }
  return result;
}

/* ϪƼ˰ʤ롣פ֤ */
int quadruple_in_hand(int pid, int n)
{
  player_t *p = &players[pid];
  int count;
  int i;
  
  p->revealed[p->revealed_count++] = toggle_revealed_bit(make_set(n, QUADRUPLET));
  p->hand[n] -= 4;

  /* ̵Фα¤ǧ뤫ɤ... */
  
  for (i = 0, count = 0; i < p->revealed_count; i++) {
    if (type_of(p->revealed[i]) == QUADRUPLET)
      count++; else break;
  }
  if (count != 4 && pool_count_quadruplets() == 4)
    return -1;
  else
    return pool_draw_after_quadruplet(pid);
}
