/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: graphics.c,v 1.2 1999/11/10 12:41:35 shinmaru Exp $
 */

#include "mahjong.h"
#include "graphics.h"
#include "x11/x11.h"

#ifndef IMAGEDIR
#define IMAGEDIR "../images"
#endif

typedef struct {
  Pixmap pixmap;
  int x;
  int y;
  int width;
  int height;
} image_t;

static GC __gc;

Pixmap yagamo_icon;

static Pixmap __hand[MAX_DIRECTION];
static Pixmap __face[MAX_DIRECTION];
static Pixmap __back[2];
static Pixmap __mask1, __mask2, __mask3, __mask4, __mask5;

static Pixmap __bucket, __handrail[2];

static Pixmap __mask[MAX_SURFACE][MAX_DIRECTION];
static image_t __image[MAX_SURFACE][MAX_DIRECTION][MAX_SORT];

static Pixmap __stick[MAX_DIRECTION][MAX_STICK];
static Pixmap __yagamo[2];

static unsigned long __blue_pixel;

static rectangle_t *highlight;
static rectangle_t *screening;
static rectangle_t *selection;

static int load_image(const char *filename, Pixmap *image, Pixmap *mask)
{
  char path[256];
  int result;

  sprintf(path, IMAGEDIR "/%s", filename);

  result = XpmReadFileToPixmap(display, main_window, path, image, mask, NULL);
  if (result < 0) {
    switch (result) {
    case XpmOpenFailed:
      fprintf(stderr, "Cannot open xpm file: %s\n", path);
      break;
    case XpmFileInvalid:
      fprintf(stderr, "Xpm file may be broken: %s\n", filename);
      break;
    case XpmNoMemory:
      fprintf(stderr, "Memory exhausted.\n");
      break;
    case XpmColorFailed:
      fprintf(stderr, "Color allocation failed.\n");
      break;
    default:
      fprintf(stderr, "Unknown reason: %s\n", filename);
    }
    return 1;
  }
  return 0;
}

static int setup_images(void)
{
  struct {
    const char *filename;
    Pixmap *image;
    Pixmap *mask;
  } img_src[] = {
    { "hand_0.xpm", &__hand[0], NULL /* __mask2 */ },
    { "hand_1.xpm", &__hand[1], &__mask1 },
    { "hand_2.xpm", &__hand[2], &__mask2 },
    { "hand_3.xpm", &__hand[3], &__mask3 },
    { "face_0.xpm", &__face[0], NULL /* __mask4 */ },
    { "face_1.xpm", &__face[1], NULL /* __mask5 */ },
    { "face_2.xpm", &__face[2], NULL /* __mask4 */ },
    { "face_3.xpm", &__face[3], NULL /* __mask5 */ },
    { "back_0.xpm", &__back[0], &__mask4 },
    { "back_1.xpm", &__back[1], &__mask5 },
    { "icon.xpm", &yagamo_icon, NULL }
  };

  Pixmap *mask_src[MAX_SURFACE][MAX_DIRECTION] = {
    { &__mask2, &__mask1, &__mask2, &__mask3 }, /*  */
    { &__mask4, &__mask5, &__mask4, &__mask5 }, /* ɽ */
    { &__mask4, &__mask5, &__mask4, &__mask5 }, /*  */
  };
  int surface, d, n;
  
  for (n = 0; n < sizeof(img_src)/sizeof(img_src[0]); n++) {
    if (load_image(img_src[n].filename, img_src[n].image, img_src[n].mask))
      return 1;
  }
  for (surface = 0; surface < MAX_SURFACE; surface++) {
    for (d = 0; d < MAX_DIRECTION; d++) {
      __mask[surface][d] = *mask_src[surface][d];
    }
  }

  /*  */
  for (n = 0; n < MAX_SORT; n++) {
    image_t *img = &__image[0][0][n];

    img->pixmap = __hand[0];
    img->x = n % 9 * image_size[0][0].width;
    img->y = n / 9 * image_size[0][0].height;
    img->width = image_size[0][0].width;
    img->height = image_size[0][0].height;

    for (d = 1; d < MAX_DIRECTION; d++) {
      img = &__image[0][d][n];
      img->pixmap = __hand[d];
      img->x = 0;
      img->y = 0;
      img->width = image_size[0][d].width;
      img->height = image_size[0][d].height;
    }
  }

  /* ɽ */
  for (d = 0; d < MAX_DIRECTION; d++) {
    for (n = 0; n < MAX_SORT; n++) {
      image_t *img = &__image[1][d][n];
      
      img->pixmap = __face[d];
      img->x = n % 9 * image_size[1][d].width;
      img->y = n / 9 * image_size[1][d].height;
      img->width = image_size[1][d].width;
      img->height = image_size[1][d].height;
    }
  }

  /*  */
  for (n = 0; n < MAX_SORT; n++) {
    image_t *img = &__image[2][0][n];

    img->pixmap = __back[0];
    img->x = 0;
    img->y = 0;
    img->width = image_size[2][0].width;
    img->height = image_size[2][0].height;
    __image[2][2][n] = *img;

    img = &__image[2][1][n];

    img->pixmap = __back[1];
    img->x = 0;
    img->y = 0;
    img->width = image_size[2][1].width;
    img->height = image_size[2][1].height;
    __image[2][3][n] = *img;
  }

  /*  */
  if (load_image("hstk_0.xpm", &__stick[0][0], NULL) ||
      load_image("hstk_1.xpm", &__stick[0][1], NULL) ||
      load_image("hstk_2.xpm", &__stick[0][2], NULL) ||
      load_image("hstk_3.xpm", &__stick[0][3], NULL) ||
      load_image("vstk_0.xpm", &__stick[1][0], NULL) ||
      load_image("vstk_1.xpm", &__stick[1][1], NULL) ||
      load_image("vstk_2.xpm", &__stick[1][2], NULL) ||
      load_image("vstk_3.xpm", &__stick[1][3], NULL)) {
    return 1;
  }
  for (n = 0; n < MAX_STICK; n++) {
    __stick[2][n] = __stick[0][n];
    __stick[3][n] = __stick[1][n];
  }

  /* ȥޡ */
  if (load_image("yagamo1.xpm", &__yagamo[0], NULL) ||
      load_image("yagamo2.xpm", &__yagamo[1], NULL)) {
    return 1;
  }
  return 0;
}

int setup_graphics(void)
{
  XGCValues v;
  int i, depth;

  if (setup_images()) {
    return 1;
  }

  depth = DefaultDepth(display, screen);

  for (i = 0; i < 2; i++) {
    const rectangle_t *r = &layout_hand[i];
    __handrail[i] = XCreatePixmap(display, main_window, r->width, r->height, depth);
  }

  /* ΤΰѤΥХåեϤ줾줬Ĥͭ */
  __bucket = XCreatePixmap(display, main_window,
			   layout_bucket[0].width, layout_bucket[1].height, depth);

  v.foreground = background_pixel;
  v.graphics_exposures = False;
  __gc = XCreateGC(display, main_window, GCForeground | GCGraphicsExposures, &v);

  __blue_pixel = get_pixel(0xffff, 0xffff, 0);
  return 0;
}

static void fill_rectangle(rectangle_t *r)
{
  XSetClipMask(display, __gc, None);
  XFillRectangle(display, main_window, __gc, r->x, r->y, r->width, r->height);
}

static void draw_tile(Drawable d, int x, int y, int tid)
{
  image_t *img = &__image[surface_of(tid)][direction_of(tid)][tile_of(tid)];
  Pixmap mask = __mask[surface_of(tid)][direction_of(tid)];

  XSetClipMask(display, __gc, mask);
  XSetClipOrigin(display, __gc, x, y);

  XCopyArea(display, img->pixmap, d, __gc, img->x, img->y, img->width, img->height, x, y);
}

void put_tile(int x, int y, int tid)
{
  draw_tile(main_window, x, y, tid);
}

static void draw_bones(Drawable d, const rectangle_t *r, bone_t *b)
{
  for (; b != NULL; b = b->next) {
    if (!is_disappear(b->tid))
      draw_tile(d, b->x - r->x, b->y - r->y, b->tid);
  }
}

static void draw_bones_reverse(Drawable d, const rectangle_t *r, bone_t *b)
{
  if (b != NULL) {
    draw_bones_reverse(d, r, b->next);
    if (!is_disappear(b->tid))
      draw_tile(d, b->x - r->x, b->y - r->y, b->tid);
  }
}

void repaint_handrail(int pid)
{
  const rectangle_t *r = &layout_hand[pid];
  Drawable d = __handrail[pid % 2];

  XSetClipMask(display, __gc, None);
  XFillRectangle(display, d, __gc, 0, 0, r->width, r->height);

  switch (pid) {
  case 0:
    draw_bones(d, r, players[0].exposure);
    draw_bones(d, r, players[0].handrail.bones);
    break;
  case 1:
    draw_bones_reverse(d, r, players[1].exposure);
    draw_bones(d, r, players[1].handrail.bones);
    break;
  case 2:
    draw_bones_reverse(d, r, players[2].exposure);
    draw_bones(d, r, players[2].handrail.bones);
    break;
  case 3:
    draw_bones(d, r, players[3].exposure);
    draw_bones_reverse(d, r, players[3].handrail.bones);
    break;
  }

  XCopyArea(display, d, main_window, default_gc, 0, 0, r->width, r->height, r->x, r->y);
  XFlush(display);
}

void repaint_bucket(int pid)
{
  const rectangle_t *r = &layout_bucket[pid];
  Drawable d = __bucket;

  XSetClipMask(display, __gc, None);
  XFillRectangle(display, d, __gc, 0, 0, r->width, r->height);

  switch (pid) {
  case 0: draw_bones_reverse(d, r, players[0].bucket); break;
  case 1: draw_bones(d, r, players[1].bucket); break;
  case 2: draw_bones(d, r, players[2].bucket); break;
  case 3: draw_bones_reverse(d, r, players[3].bucket); break;
  }

  XCopyArea(display, d, main_window, default_gc, 0, 0, r->width, r->height, r->x, r->y);

  if (players[pid].condition & (bit(COND_READY) | bit(COND_ONLY_DECLARED))) {
    r = &layout_ready[pid];
    XCopyArea(display, __stick[pid][1], main_window, default_gc, 0, 0,
	      r->width, r->height, r->x, r->y);
  }
  XFlush(display);
}

void repaint_wall(void)
{
  const rectangle_t *r = &layout_wall;
  Drawable d = __bucket;

  XSetClipMask(display, __gc, None);
  XFillRectangle(display, d, __gc, 0, 0, r->width, r->height);

  draw_bones_reverse(d, r, wall_bones);

  XCopyArea(display, d, main_window, default_gc, 0, 0, r->width, r->height, r->x, r->y);
}

static void repaint_yagamo(void)
{
  int pid = roundkeeper.first_dealer;

  const rectangle_t *r = &layout_yagamo[pid];
  Drawable d = __yagamo[min(1, roundkeeper.round_wind)];

  XCopyArea(display, d, main_window, default_gc, 0, pid * r->height,
	    r->width, r->height, r->x, r->y);
}

void repaint_all(void)
{
  int i;

  repaint_yagamo();

  repaint_wall();

  for (i = 0; i < MAX_PLAYER; i++) {
    repaint_bucket(i);
    repaint_handrail(i);
  }

  set_highlight(highlight);
  set_screening(screening);
  set_selection(selection);

  XFlush(display);
}

void set_highlight(rectangle_t *r)
{
  XSetForeground(display, __gc, __blue_pixel);
  XSetFunction(display, __gc, GXxor);
  if (highlight) fill_rectangle(highlight);
  if (highlight != r) {
    highlight = r;
    if (highlight) fill_rectangle(highlight);
  }
  XSetForeground(display, __gc, background_pixel);
  XSetFunction(display, __gc, GXcopy);
}

void set_screening(rectangle_t *r)
{
  XSetFunction(display, __gc, GXinvert);
  if (screening) fill_rectangle(screening);
  if (screening != r) {
    screening = r;
    if (screening) fill_rectangle(screening);
  }
  XSetFunction(display, __gc, GXcopy);
}

void set_selection(rectangle_t *r)
{
  XSetFunction(display, __gc, GXxor);
  if (selection) fill_rectangle(selection);
  selection = r;
  if (selection) fill_rectangle(selection);
  XSetFunction(display, __gc, GXcopy);
}

void show_selection(void)
{
  Window root, window;
  int root_x, root_y, x, y;
  unsigned mask;

  XQueryPointer(display, main_window, &root, &window, &root_x, &root_y, &x, &y, &mask);
  if (window == main_window) {
    pointer_move(x, y);
  }
}

void clear_rectangle(int x, int y, int width, int height)
{
  XClearArea(display, main_window, x, y, width, height, False);
}

void beep(void)
{
  XBell(display, 0);
}

void draw_text(Drawable d, GC gc, int x, int y, const char *text)
{
  const char *top, *end;

  top = end = text;
  for (;;) {
    int c = *end++;
    if (c == '\0' || c == '\n') {
      int bytes = end - top -1;
      if (bytes > 0) {
	XmbDrawString(display, d, font_set, gc, x, y, top, bytes);
      }
      y += 20;
      top = end;
    }
    if (c == '\0')
      break;
  }
}

static const char *__text;
static const int *__bill;

static void paint_text(Drawable d, GC gc)
{
  XFillRectangle(display, d, gc, 16, 36, 256, 280);

  XSetForeground(display, gc, BlackPixel(display, screen));

  XmbDrawString(display, d, font_set, gc, 16, 16 - font_size.y,
		round_caption, strlen(round_caption));

  XSetForeground(display, gc, WhitePixel(display, screen));

  draw_text(d, gc, 24, 36 - font_size.y, __text);
}

static void paint_bill(Drawable d, GC gc)
{
  static const rectangle_t r[MAX_PLAYER] = {
    {  88, 228,  112, 80 },
    { 152, 136,  112, 80 },
    {  88,  44,  112, 80 },
    {  24, 136,  112, 80 }
  };
  char buf[256];
  int x, y;
  int i;

  XFillRectangle(display, d, gc, 16, 36, 256, 280);

  XSetForeground(display, gc, WhitePixel(display, screen));

  for (i = 0; i < MAX_PLAYER; i++) {
    XFillRectangle(display, d, gc, r[i].x, r[i].y, r[i].width, r[i].height);
  }

  XSetForeground(display, gc, BlackPixel(display, screen));

  y = 16 - font_size.y;
  XmbDrawString(display, d, font_set, gc, 16, y,
		round_caption, strlen(round_caption));

  for (i = 0; i < MAX_PLAYER; i++) {
    int length;

    x = r[i].x;
    y = r[i].y;

    length = sprintf(buf, "%s", tile_face[seat_wind_of(i)]);
    y -= font_size.y;
    XmbDrawString(display, d, font_set, gc, x, y, buf, length);

    length = sprintf(buf, "%12d", player_score[i]);
    y += 20;
    XmbDrawString(display, d, font_set, gc, x, y, buf, length);

    length = sprintf(buf, __bill[i] != 0 ? "%+12d" : "%12d", __bill[i]);
    y += 20;
    XmbDrawString(display, d, font_set, gc, x, y, buf, length);

    length = sprintf(buf, "%12d", player_score[i] + __bill[i]);
    y += 20;
    XmbDrawString(display, d, font_set, gc, x, y, buf, length);
  }
}

void report_dialog(const char *text)
{
  if (text != NULL) {
    __text = text;
    show_modal_dialog(paint_text);
  }
}

void score_dialog(const int *bill)
{
  if (bill != NULL) {
    __bill = bill;
    show_modal_dialog(paint_bill);
  }
}

void open_hand(int pid)
{
  bone_t *b;
  for (b = players[pid].handrail.bones; b != NULL; b = b->next) {
    if (surface_of(b->tid) != SURFACE_FACE)
      toggle_handrail_surface(pid, b, SURFACE_FACE);
  }
  repaint_handrail(pid);
}

void open_wall(void)
{
  bone_t *b;

  for (b = wall_bones; b != NULL; b = b->next) {
    if (surface_of(b->tid) == SURFACE_FACE && b->next != NULL) {
      b->y += 48;
      b = b->next;
      b->tid = make_tid(tile_of(b->tid), 0, SURFACE_FACE);
    }
  }
  repaint_wall();
}
