/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: main.c,v 1.3 1999/11/11 07:58:46 shinmaru Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include "mahjong.h"
#include "option.h"
#include "graphics.h"
#include "x11/x11.h"

Display *display;
int screen;
GC default_gc;
Window main_window;
unsigned long background_pixel;

extern const char *DEFAULT_FONT_LIST;

XFontSet font_set;
XRectangle font_size;

static Window dialog;
static unsigned long dialog_pixel;

static Atom WM_PROTOCOLS, WM_DELETE_WINDOW;

static void version_dialog(void);

unsigned long get_pixel(int red, int green, int blue)
{
  XColor c;

  c.red = red;
  c.green = green; 
  c.blue = blue;

  XAllocColor(display, DefaultColormap(display, screen), &c);

  return c.pixel;
}

static int create_font(const char *font_list)
{
  char **missing_list;
  int missing_count;
  char *def_string;

  font_set = XCreateFontSet(display, font_list, &missing_list, &missing_count,
			    &def_string);
  if (font_set == NULL) {
    int i;

    fprintf(stderr, "Can't create font set.\n");
    for (i = 0; i < missing_count; i++) {
      fprintf(stderr, " missing: %s\n", missing_list[i]);
    }
    return 1;
  }
  XFreeStringList(missing_list);

  font_size = XExtentsOfFontSet(font_set)->max_logical_extent;
  return 0;
}

static int close_display(void)
{
  XCloseDisplay(display);
  return 0;
}

int main(int argc, char *argv[])
{
  XSizeHints hint;
  XEvent e;

  setlocale(LC_CTYPE, "");

  option_silence = 1;
  option_wait = 1;
  option_batch = 0;
  option_stupid = 0;

  parse_args(argc, argv);

  srandom(random_seed);
  fprintf(stderr, "%x\n", random_seed);

  display = XOpenDisplay(NULL);
  if (display == NULL) {
    fprintf(stderr, "Can't open display.\n");
    return 1;
  }

  WM_PROTOCOLS = XInternAtom(display, "WM_PROTOCOLS", False);
  WM_DELETE_WINDOW = XInternAtom(display, "WM_DELETE_WINDOW", False);

  screen = DefaultScreen(display);

  default_gc = DefaultGC(display, screen);

  background_pixel = get_pixel(0, 0x8000, 0);

  main_window = XCreateSimpleWindow(display, RootWindow(display, screen),
				    0, 0, layout_deck.width, layout_deck.height,
				    1,
				    WhitePixel(display, screen),
				    background_pixel);

  XSelectInput(display, main_window, KeyPressMask | ButtonPressMask | ExposureMask);

  hint.width = hint.min_width = hint.max_width = layout_deck.width;
  hint.height = hint.min_height = hint.max_height = layout_deck.height;
  hint.flags = PSize | PMinSize | PMaxSize;
  XSetWMNormalHints(display, main_window, &hint);

  XStoreName(display, main_window, "xyagamo");
  XSetWMProtocols(display, main_window, &WM_DELETE_WINDOW, 1);

  dialog_pixel = get_pixel(0, 0, 0x8000);

  dialog = XCreateSimpleWindow(display, main_window,
			       128, 128, 288, 328,
			       1,
			       WhitePixel(display, screen),
			       WhitePixel(display, screen));
  XSelectInput(display, dialog, ButtonPressMask | ExposureMask);

  if (create_font(DEFAULT_FONT_LIST)) return 1;

  if (setup_graphics()) return 1;

  XMapWindow(display, main_window);
  for (;;) {
    XNextEvent(display, &e);
    if (e.type == Expose)
      break;
  }

  start_round(random_integer(4));

  for (;;) {
    XNextEvent(display, &e);

    switch (e.type) {
    case KeyPress:
      {
	KeySym keysym = XKeycodeToKeysym(display, e.xkey.keycode, 0);
	if (keysym == 'q' && (e.xkey.state & ControlMask) != 0) {
	  exit(close_display());
	}
	if (keysym == 'v') {
	  version_dialog();
	}
      }
      break;

    case ButtonPress:
      if (round_ended()) {
	version_dialog();
	XClearWindow(display, main_window);
	start_round(random_integer(4));
      }
      else if (e.xbutton.button == 1) {
	if ((e.xbutton.state & ControlMask) == 0)
	  button1_down(e.xbutton.x, e.xbutton.y);
	else
	  button2_down(e.xbutton.x, e.xbutton.y);
      }
      else if (e.xbutton.button == 2) {
	button2_down(e.xbutton.x, e.xbutton.y);
      }
      else if (e.xbutton.button == 3) {
	button3_down(e.xbutton.x, e.xbutton.y);
      }
      break;

#if 0 /* I seem it may be too expensive for some X servers. */
    case MotionNotify:
      pointer_move(e.xmotion.x, e.xmotion.y);
      break;
#endif

    case Expose:
      repaint_all();
      break;

    case ClientMessage:
      if (e.xclient.message_type == WM_PROTOCOLS &&
	  e.xclient.data.l[0] == WM_DELETE_WINDOW) {
	exit(close_display());
      }
      break;
    }
  }
  return 0;
}

void show_modal_dialog(void (*paint_content)(Drawable d, GC gc))
{
  XEvent e;
  GC gc;
  XGCValues v;

  v.graphics_exposures = False;
  gc = XCreateGC(display, dialog, GCGraphicsExposures, &v);
  
  XMapWindow(display, dialog);

  for (;;) {
    XNextEvent(display, &e);

    switch (e.type) {
    case ButtonPress:
      if (e.xbutton.window == main_window) {
	XMapWindow(display, dialog);
      }
      else if (e.xbutton.window == dialog) {
	if (e.xbutton.button == 1) {
	  XUnmapWindow(display, dialog);
	  XFreeGC(display, gc);
	  return;
	}
	else if (e.xbutton.button == 3) {
	  XUnmapWindow(display, dialog);
	}
      }
      break;

    case Expose:
      if (e.xexpose.window == main_window) {
	repaint_all();
      }
      else if (e.xexpose.window == dialog) {
	XSetForeground(display, gc, dialog_pixel);
	(*paint_content)(dialog, gc);
      }
      break;

    case ClientMessage:
      if (e.xclient.message_type == WM_PROTOCOLS &&
	  e.xclient.data.l[0] == WM_DELETE_WINDOW) {
	exit(close_display());
      }
    }
  }
}

static const char *version_text =
"xyagamo 0.1b\n"
"\n"
"ײIcons\n"
"by ѻůϺ (̵Ԥβ)\n";

static const char *copyright =
"Copyright (c) 1999\n"
"Ŀ <shinmaru@dokidoki.ne.jp>\n"
"All rights reserved.\n";

void version_dialog(void)
{
  Window window;
  XEvent e;

  window = XCreateSimpleWindow(display, main_window, 111, 182, 352, 220, 1,
			       WhitePixel(display, screen),
			       WhitePixel(display, screen));

  XSelectInput(display, window, KeyPressMask | ButtonPressMask | ExposureMask);
  XMapWindow(display, window);

  for (;;) {
    XNextEvent(display, &e);

    switch (e.type) {
    case KeyPress:
    case ButtonPress:
      XDestroyWindow(display, window);
      return;

    case Expose:
      if (e.xexpose.window == main_window) {
	repaint_all();
      }
      else if (e.xexpose.window == window) {
	XCopyArea(display, yagamo_icon, window, default_gc, 0, 0, 90, 90, 16, 16);
	draw_text(window, default_gc, 128, 16 - font_size.y, version_text);
	draw_text(window, default_gc, 16, 128 - font_size.y, copyright);
      }
      break;
    }
  }
}
